#include <iostream.h>
extern "C" {
#include <string.h>
}

class Address 
{ 
public: 
  virtual const char* type() const = 0; 
  virtual void print(int international=0) const = 0; 
  virtual void acquire(void) = 0; 
  virtual ~Address();  
};

class  MailingAddress : public Address
{
public:
  virtual ~MailingAddress();
  const char* type() const;
protected:
  MailingAddress();

  void name(const char*); // set
  const char* name() const; // get

  void street(const char*); // set
  const char* street() const; // get

  void number(const char*); // set
  const char* number() const; // get

  void city(const char*); // set
  const char* city() const; // get

  void postalCode(const char*); // set
  const char* postalCode() const; // get

  void country(const char*); // set
  const char* country() const; // get
private:
  char* name_data;
  char* street_data;
  char* number_data;
  char* city_data;
  char* postalCode_data;
  char* country_data;
  //
  // declared private to disallow them
  //
  MailingAddress(const MailingAddress&);
  MailingAddress& operator=(const MailingAddress&);
};

class SwedishAddress : public MailingAddress
{
public:
  SwedishAddress();
  virtual void print(int international=0) const;
  virtual void acquire(void);
};


class USAddress : public MailingAddress
{
public:
  USAddress();
  virtual void print(int international=0) const;
  virtual void acquire(void);
};

Address::~Address()
{
}

MailingAddress::~MailingAddress()
{
  delete[] name_data;
  delete[] street_data;
  delete[] number_data;
  delete[] city_data;
  delete[] postalCode_data;
  delete[] country_data;
}


MailingAddress::MailingAddress()
  : name_data(0),
    street_data(0),
    number_data(0),
    city_data(0),
    postalCode_data(0),
    country_data(0)
{
}


const char* MailingAddress::type(void) const
{
  return "Mailing address";
}

const char* MailingAddress::name(void) const
{
  return name_data;
}

const char* MailingAddress::street(void) const
{
  return street_data;
}

const char* MailingAddress::number(void) const
{
  return number_data;
}

const char* MailingAddress::city(void) const
{
  return city_data;
}

const char* MailingAddress::postalCode(void) const
{
  return postalCode_data;
}

const char* MailingAddress::country(void) const
{
  return country_data;
}


static void replace(char*& data, const char* n)
{
  if (data != n) {
    delete[] data;
    data = new char[strlen(n)+1];
    strcpy(data,n);
  }
}


void MailingAddress::name(const char* n)
{
  ::replace(name_data,n);
}

void MailingAddress::street(const char* n)
{
  ::replace(street_data,n);
}

void MailingAddress::number(const char* n)
{
  ::replace(number_data,n);
}

void MailingAddress::city(const char* n)
{
  ::replace(city_data,n);
}

void MailingAddress::postalCode(const char* n)
{
  ::replace(postalCode_data,n);
}

void MailingAddress::country(const char* n)
{
  ::replace(country_data,n);
}


SwedishAddress::SwedishAddress()
  : MailingAddress()
{
  country("Sweden"); // what else?
}

void SwedishAddress::print(int international) const
{
  cout << name() << endl;
  cout << street() << ' ' << number() << endl;
  cout << postalCode() << ' ' << city() << endl;
  if (international) cout << country() << endl;
}

void SwedishAddress::acquire(void)
{
  char buffer[100]; // A mighty long field

  cout << "Name: " << flush;
  cin.getline(buffer,sizeof(buffer));
  name(buffer);

  cout << "Street: " << flush;
  cin.getline(buffer,sizeof(buffer));
  street(buffer);

  cout << "Number: " << flush;
  cin.getline(buffer,sizeof(buffer));
  number(buffer);

  cout << "Postal code: " << flush;
  cin.getline(buffer,sizeof(buffer));
  postalCode(buffer);

  cout << "City: " << flush;
  cin.getline(buffer,sizeof(buffer));
  city(buffer);
}

USAddress::USAddress()
  : MailingAddress()
{
  country("U.S.A."); // what else?
}

void USAddress::print(int international) const
{
  cout << name() << endl;
  cout << number() << ' ' << street() << endl;
  cout << city() << ' ' << postalCode() << endl;
  if (international) cout << country() << endl;
}

void USAddress::acquire(void)
{
  char buffer[100]; // Seems like a mighty long field

  cout << "Name: " << flush;
  cin.getline(buffer,sizeof(buffer));
  name(buffer);

  cout << "Number: " << flush;
  cin.getline(buffer,sizeof(buffer));
  number(buffer);

  cout << "Street: " << flush;
  cin.getline( buffer,sizeof(buffer));
  street(buffer);

  cout << "City: " << flush;
  cin.getline(buffer, sizeof(buffer));
  city(buffer);

  cout << "State and ZIP: " << flush;
  cin.getline(buffer,sizeof(buffer));
  postalCode(buffer);
}


Address** get_addrs(Address** first,Address** last)
{
  Address** current = first;
  while (current != last)
  {
    cout << endl << "Kind (U)S, (S)wedish or (N)one "
	 << flush;

    char answer[5]; // Should be enough.
    cin.getline(answer,sizeof(answer));
    if (!cin) break;

    switch (answer[0]) {
    case 'U': case 'u':
      *current = new USAddress;
      break;
    case 'S': case 's':
      *current = new SwedishAddress;
      break;
    default:
      return current;
    }
    (**current).acquire();
    ++current;
  }
  return current;
}



template <class OI,class F>
void for_each(OI first,OI last, const F& functor)
{
  while (first != last) {
    functor(*first);
    ++first;
  }
}

class print
{
public:
  print(int i) ;
  void operator()(const Address*) const;
private:
  int international;
};

print::print(int i)
  : international(i)
{
}

void print::operator()(const Address* p) const
{
  p->print(international);
  cout << endl;
}

template <class T>
class deallocate
{
public:
  void operator()(T* p) const; 
};

template <class T>
void deallocate<T>::operator()(T* p) const
{ 
  delete p;
}


int main(void)
{
  const unsigned size=10;
  Address* addrs[size];
  Address** first = addrs; // needed for VACPP (bug?)
  Address** last = get_addrs(addrs,addrs+size);

  cout << endl << "--------" << endl;

  for_each(first,last,print(1));
  for_each(first,last,deallocate<Address>());
  return 0;
}
